/*
 * Copyright (C) 2008 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include "loudgame.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "loa-board.h"

typedef struct _loa_game {
    loudgame_t lg;
    loa_board_t board;
} loa_game_t;

static void
loa_game_new_game (loa_game_t *game)
{
    loa_board_reset (&game->board);
}

static loa_bool_t
loa_game_move (loa_game_t *game, const char * peer,
	       int x1, int y1, int x2, int y2)
{
    char *error;

    if (! loa_board_move (&game->board, x1, y1, x2, y2, &error)) {
	loudgame_sendf (&game->lg, peer, "Illegal move: %c%d%c%d: %s",
			'a' + x1, LOA_BOARD_SIZE - y1,
			'a' + x2, LOA_BOARD_SIZE - y2,
			error);
	return FALSE;
    }

    return TRUE;
}

static void
loa_game_handle_show (loa_game_t *game,
		      const char *peer)
{
    LmMessage *reply;
    LmMessageNode *html, *body, *span;
    gboolean result;
    GError *error = NULL;
    char *board_string;
    char *line, *newline;

    reply = lm_message_new (peer, LM_MESSAGE_TYPE_MESSAGE);
    html = lm_message_node_add_child (reply->node, "html", "");
    lm_message_node_set_attribute (html,
				   "xmlns",
				   "http://jabber.org/protocol/xhtml-im");
    body = lm_message_node_add_child (html, "body", "");
    lm_message_node_set_attribute (body,
				   "xmlns",
				   "http://www.w3.org/1999/xhtml");

    if (game->board.player == LOA_PLAYER_BLACK)
	lm_message_node_add_child (body, "span", "Black to move:");
    else
	lm_message_node_add_child (body, "span", "White to move:");

    board_string = loa_board_to_string (&game->board);

    line = board_string;
    while (1) {
	newline = strchr (line, '\n');
	if (newline)
	    *newline = '\0';
	lm_message_node_add_child (body, "br", "");
	span = lm_message_node_add_child (body, "span", line);
	lm_message_node_set_attribute (span, "style",
				       "font-family: Monospace;");
	if (! newline)
	    break;
	line = newline + 1;
    }

    free (board_string);

    result = lm_connection_send (game->lg.connection, reply, &error);
    lm_message_unref (reply);

    if (! result) {
	g_error ("lm_connection_send failed: %s\n",
		 error->message);
	loudgame_quit (&game->lg, 1);
    }
}

static void
loa_game_handle_move (loa_game_t *game,
		      const char *peer,
		      const char *move)
{
    char xc1, xc2;
    int x1, y1, x2, y2;
    int matched;

    matched = sscanf (move, " %c %d %c %d ", &xc1, &y1, &xc2, &y2);
    if (matched != 4) {
	loudgame_sendf (&game->lg, peer,
			"Error: The 'move' command requires a move of the form 'b1d3'");
	return;
    }

    x1 = tolower (xc1) - 'a';
    x2 = tolower (xc2) - 'a';
    /* We use an upper-left origin internally. */
    y1 = LOA_BOARD_SIZE - y1;
    y2 = LOA_BOARD_SIZE - y2;
    if (! loa_game_move (game, peer, x1, y1, x2, y2))
	return;

    loudgame_broadcastf (&game->lg, "%c%d%c%d",
			 'a' + x1, LOA_BOARD_SIZE - y1,
			 'a' + x2, LOA_BOARD_SIZE - y2);

    if (loa_board_is_won (&game->board, x2, y2))
	loudgame_broadcastf (&game->lg, "%s wins", peer);
}

static void
loa_game_handle_history (loa_game_t *game,
			 const char *peer)
{
    int i;

    for (i = 0; i < game->board.num_moves; i++)
	loudgame_sendf (&game->lg, peer, "%s",
			loa_move_to_string (&game->board.moves[i]));
}

static void
loa_game_handle_pass (loa_game_t *game, const char *peer)
{
    loudgame_broadcastf (&game->lg, "%s passes", peer);

    if (loa_board_pass (&game->board))
	return;

    loudgame_sendf (&game->lg, peer,
		    "Error: You cannot pass since you have a legal move available");
}

static void
loa_game_handle_help (loa_game_t *game, const char *peer)
{
    loudgame_sendf (&game->lg, peer,
		    "I'm a bot that allows you to play the game Lines of Action.\n"
		    "Here are some generic commands I understand:\n"
		    LOUDGAME_HELP
		    "\n"
		    "And some game-specific commands:\n"
		    "\tshow     \t\tShow the current board\n"
		    "\tmove aNbN\tMove a piece, (eg. 'move b1d3')\n"
		    "\tpass     \t\tSkip a turn (only legal if no moves are possible)\n"
		    "\tnew      \t\tBegin a new game\n"
		    "\thistory  \t\tShow the move history of the game\n"
		    "\thelp     \t\tThis help message\n"
		    "\trules    \t\tA description of the Lines of Action rules\n"
		    "\n"
		    "Lines of Action was invented by Claude Soucie and first made popular\n"
		    "when its rules were published by Sid Sackson in \"A Gamut of Games\" (1969).\n"
		    "\n"
		    "If you are new to Lines of Action, type 'rules' now to learn the rules.\n");
}

static void
loa_game_handle_rules (loa_game_t *game, const char *peer)
{
    loudgame_sendf (&game->lg, peer,
		    "Lines of Action can be played with a standard (English) checkers set,\n"
		    "that is an 8x8 board and 12 markers each of contrasting colors ('black' and\n"
		    "'white'). The initial placement has 6 each of the black pieces on the top\n"
		    "and bottom rows, and 6 each of the white pieces on the left and right columns\n"
		    "leaving the four corner spaces empty. Play begins with the black player and\n"
		    "then alternates.\n"
		    "\n"
		    "On each move a piece is moved in a straight line in any of eight directions,\n"
		    "(similar to a queen's move in chess), but must be moved exactly the same\n"
		    "number of spaces as there are pieces (of either color) in the row, column,\n"
		    "or diagonal of the move. A piece may jump over pieces of its own color, but\n"
		    "may not jump a piece of the opposite color. The final square of the move can\n"
		    "be either empty or can contain a piece of the opposing color, in which case\n"
		    "that piece is removed from the game.\n"
		    "\n"
		    "If a player has no possible move, then that player must pass, (but if the\n"
		    "player has a possible move, then the player cannot pass).\n"
		    "\n"
		    "The goal of the game is to connect all of your remaining pieces into\n"
		    "a single, connected group. Pieces that are diagonally adjacent are\n"
		    "considered connected.\n"
		    "\n"
		    "If a move simultaneously creates a winning condition for both players, this\n"
		    "is considered a win for the player making the move.\n"
		    "\n"
		    "Notes on this implementation:\n"
		    "\n"
		    "This implementation enforces the move rules described above, but will\n"
		    "allow any person to make moves, (that is, you can control both\n"
		    "colors). Think of it as being somewhat like a physical board where\n"
		    "any person could move the pieces at any time, but generally two people\n"
		    "will take turns and others will politely watch.\n"
		    "\n"
		    "Also, the game won't currently show you the board again after a\n"
		    "player makes a move, so you'll need to keep typing 'show' to see\n"
		    "updates. Yes, this is a bug, and yes it will be fixed soon.\n"
		    "\n"
		    "Lines of Action was invented by Claude Soucie and first made popular\n"
		    "when its rules were published by Sid Sackson in \"A Gamut of Games\" (1969).\n");
}

static void
loa_game_handle_new (loa_game_t *game, const char *peer)
{
    loudgame_broadcastf (&game->lg, "%s has started a new game", peer);
    loa_game_new_game (game);
}

static void
loa_game_handle_message (loudgame_t *lg,
			 const char *peer,
			 const char *message)
{
    loa_game_t *game = (loa_game_t *) lg;

    if (strcmp (message, "show") == 0)
	loa_game_handle_show (game, peer);
    else if (strncmp (message, "move", 4) == 0)
	loa_game_handle_move (game, peer, message + 4);
    else if (strcmp (message, "pass") == 0)
	loa_game_handle_pass (game, peer);
    else if (strcmp (message, "new") == 0)
	loa_game_handle_new (game, peer);
    else if (strcmp (message, "history") == 0)
	loa_game_handle_history (game, peer);
    else if (strcmp (message, "help") == 0)
	loa_game_handle_help (game, peer);
    else if (strcmp (message, "rules") == 0)
	loa_game_handle_rules (game, peer);
    else
	loudgame_sendf (lg, peer, "Unknown command: '%s'. Use 'help' to get a list of valid commands.", message);
}

static int
loa_game_init (loa_game_t *game, int argc, char *argv[])
{
    int err;

    err = loudgame_init (&game->lg, argc, argv);
    if (err)
	return err;

    loa_board_init (&game->board);

    loa_game_new_game (game);

    return 0;
}

int
main (int argc, char **argv)
{
    loa_game_t game;
    int err;

    err = loa_game_init (&game, argc, argv);
    if (err)
	return err;

    game.lg.handle_message = loa_game_handle_message;

    err = loudgame_run (&game.lg);
    if (err)
	return err;

    return 0;
}
